import pandas as pd
import nltk
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import word_tokenize
from colour import Color
from difflib import get_close_matches

nltk.download('wordnet')
nltk.download('omw-1.4')
import re
from word2number import w2n
import yaml

class Helper:

    def get_key_using_val(self, dic1, val1):

        # getting the kay name from dic1 that has a value val1
        key_list = list(dic1.keys())
        val_list = list(dic1.values())
        position = val_list.index(val1)
        #         key_name =  key_list[position]
        #         dic2[key_name] = dic2.pop('color')
        return key_list[position]

    def _update_dictionary(self, dict1, dict2):
        dict1 = {} if not dict1 else dict1
        dict2 = {} if not dict2 else dict2
        if dict2 and dict1:
            dict2.update(dict1)
            return dict2

        elif dict2:
            return dict2
        else:
            return dict1

    def seperate_digits(self, tokens):
        for i in range(len(tokens)):
            if not self.is_num(tokens[i]):
                splitted = re.split(r'(\d+)', tokens[i])
                splitted = [s for s in splitted if s not in ['']]
                joined = ''
                for j in range(len(splitted)):
                    if j >= len(splitted) - 1:
                        joined += splitted[j]
                    elif splitted[j + 1][0].isalpha() and splitted[j + 1].lower() != 'x':
                        joined += splitted[j] + ' '
                    else:
                        joined += splitted[j]
                tokens[i] = joined.replace('. ', '.').replace(' .', '.').lower()
        return tokens

    def check_digit(self, text):

        try:
            w2n.word_to_num(text)
            return True
        except ValueError:
            return False

    def get_digits(self, tokens):
        digits = [re.findall(r'\d+', token)[0] for token in tokens if re.findall(r'\d+', token)]
        if not digits:
            digits = [str(w2n.word_to_num(token)) for token in tokens if self.check_digit(token)]
        return digits

    def get_dimensions(self, value):
        units = ['cm', 'inch']
        unit = [unit for unit in units if unit in value.lower()]
        try:
            unit = unit[0]
        except:
            unit = ''
        value = value.replace(',', '.')
        d = value.lower().split('x')
        if len(d) < 2:
            d = value.lower().split(' ')
        digits = []
        for index, i in enumerate(d):
            try:
                num = digits.append(re.search(r'[0-9\.]+', i).group())
            except:
                pass

            if index > 2:
                continue
        digits = [digit for digit in digits if digit not in ['.', '', ' ']][0:3]
        if not digits:
            return value

        return " X ".join([digit for digit in digits]) + ' ' + unit

    def is_num(self, text):
        try:
            float(text)
            return True
        except:
            return False

    def map_tokens(self, map_dict, feature_name, tokens):
        for i in range(len(tokens)):
            if map_dict.get(feature_name) and map_dict.get(feature_name).get(tokens[i].lower()):
                tokens[i] = map_dict.get(feature_name).get(tokens[i].lower())
            elif map_dict.get(tokens[i].lower()):
                tokens[i] = map_dict.get(tokens[i].lower())
        return tokens

    def map_value(self, map_dict, feature_name, value):

        if map_dict.get(feature_name) and map_dict.get(feature_name).get(value.lower()):
            value = map_dict.get(feature_name).get(value.lower()).title()
        elif map_dict.get(value.lower()):
            value = map_dict.get(value.lower()).title()
        return value


class DataQualityTransformer(Helper):

    def __init__(self, tag='clothing'):

        with open(r'configs_all.yml') as file:
            # The FullLoader parameter handles the conversion from YAML
            # scalar values to Python the dictionary format
            config = yaml.load(file, Loader=yaml.FullLoader)

        standard_values = config['standard_values']
        exclude_set = config['exclude_set']
        feature_group = config['feature_group']
        no_join_features = config['no_join_features']
        con_map = config['con_map']
        value_map = config['value_map']
        skip_set = config['skip_set']
        additional_colors = config['additional_colors']
        small_words = config['small_words']
        capital_words = config['capital_words']
        plural = config['plural']
        numeric_features = config['numeric_features']
        sentence = config['sentence']

        self.tag = tag
        # Lemmatizer
        self.wnl = WordNetLemmatizer()

        self.is_noun = lambda pos: pos[0] == 'N'
        self.is_verb = lambda pos: pos[0] == 'V'
        self.is_adjective = lambda pos: pos[0] == 'J'
        self.is_adverb = lambda pos: pos[0] == 'R'
        self.is_conjunction = lambda pos: pos == 'CC'
        self.excluded = lambda pos: pos == 'PDT'
        self.has_cd = lambda poss: any([pos[1] == 'CD' for pos in poss])

        self.WN_NOUN = 'n'
        self.WN_VERB = 'v'
        self.WN_ADJECTIVE = 'a'
        self.WN_ADJECTIVE_SATELLITE = 's'
        self.WN_ADVERB = 'r'

        self.feature_parser = {'default': self.feature_value_parser}
        self.no_join_features = no_join_features
        self.additional_colors = additional_colors
        self.numeric_features = numeric_features.get(tag, [])
        self.sentence = self._update_dictionary(sentence.get('default', []), sentence.get(tag, []))

        # works only on one token at a time with adding the default values
        self.feature_group = self._update_dictionary(feature_group.get('default', []), feature_group.get(tag, []))
        self.skip_set = self._update_dictionary(skip_set.get('default', []), skip_set.get(tag, []))
        self.exclude_set = self._update_dictionary(exclude_set.get('default', []), exclude_set.get(tag, []))
        self.plural = self._update_dictionary(plural.get('default', []), plural.get(tag, []))
        self.small_words = self._update_dictionary(small_words.get('default', []), small_words.get(tag, []))
        self.capital_words = self._update_dictionary(capital_words.get('default', []), capital_words.get(tag, []))

        color_feature_name = self.get_key_using_val(self.feature_group, 'color')
        # Normalization
        self.standard_values = self._update_dictionary(standard_values.get('default', []), standard_values.get(tag, []))
        self.standard_values[color_feature_name] = self.standard_values.pop('color')

        self.value_map = self._update_dictionary(value_map.get('default', []), value_map.get(tag, []))
        self.value_map[color_feature_name] = self.value_map.pop('color')

        self.con_map = self._update_dictionary(con_map.get('default', []), con_map.get(tag, []))
        self.con_map[color_feature_name] = self.con_map.pop('color')

    def feature_value_parser(self, feature_name, value):
        if self.standard_values.get(feature_name) and 'color' not in feature_name.lower():
            matches = get_close_matches(value.title(), self.standard_values[feature_name], 1, 0.5)

            if matches:
                return matches[0]
            else:
                matches = [get_close_matches(token.title(), self.standard_values[feature_name], 1, 0.65) for token in
                           word_tokenize(value)]
                matches = [match[0] for match in matches if match]
                if matches:
                    return matches[0]
        return value

    def check_color(self, color):
        try:
            if color.lower().strip() in self.additional_colors:
                return True
            Color(color)
            return True
        except ValueError:
            return False

    def get_lemma(self, word, pos, feature_name):
        lemma_word = word

        if 'color' in feature_name.lower() and not self.check_color(lemma_word.lower()):
            return ''
        if word.lower() in self.skip_set.get(feature_name, {}) or word.lower() in self.skip_set:
            return word
        if self.is_noun(pos):
            lemma_word = self.wnl.lemmatize(word.lower(), self.WN_NOUN)
            if lemma_word == word:
                lemma_word = self.wnl.lemmatize(word.lower(), self.WN_VERB)
        elif self.is_verb(pos):
            lemma_word = self.wnl.lemmatize(word.lower(), self.WN_VERB)
        elif self.is_adjective(pos):
            lemma_word = self.wnl.lemmatize(word.lower(), self.WN_ADJECTIVE)
        elif self.is_adverb(pos):
            lemma_word = self.wnl.lemmatize(word.lower(), self.WN_ADVERB)
        if lemma_word.lower() in self.exclude_set.get(feature_name, {}) or lemma_word.lower() in self.exclude_set:
            return ''
        return lemma_word

    def special_tokonizer(self, feature_name, featue_value):
        special_tokens = []
        if self.standard_values.get(feature_name):
            special_tokens = [token.lower() for token in self.standard_values.get(feature_name) if
                              token.lower() in featue_value.lower()]
        rest_of_values = re.sub("|".join(special_tokens), "", featue_value.lower()).split(" ")
        special_tokens = [token.title() for token in special_tokens if token != '']
        rest_of_values = [val.title() for val in rest_of_values if val != '']
        return special_tokens + rest_of_values

    def characters_cleaning(self, value):
        'remove any characters/ sepcial characters'
        value = value.replace('-', ' ')
        return value.replace(',', ' ')

    def tokenizing(self, feature_name, value):
        'return tokens after some processing'
        tokens = [token.lower() for token in word_tokenize(value)]
        # remove '.' for non difits
        tokens = [token.replace('.', '') if not token.replace('.', '', 1).isdigit() else token for token in tokens]
        # remove '/' for non digits and size feature name
        return [token.replace('/', ' ') if not (
                token.replace('/', '', 1).isdigit() or 'size' in feature_name.lower()) else token for token in
                tokens]

    def con_mapper(self, tokens, con_map):
        return [con_map.get(word.lower().strip(), word.strip()) for word in tokens]

    def lemmetize(self, feature_name, tokens, lemmetizer):
        words_pos = nltk.pos_tag(tokens)
        return [lemmetizer(word, pos, feature_name) for (word, pos) in words_pos if not self.excluded(pos)]

    def join_captelaize_nouns(self, tokens):
        words_pos = nltk.pos_tag(tokens)
        return ' '.join(
            [word.strip().title() if word != '' and not self.is_conjunction(pos) else word.strip() for (word, pos) in
             words_pos])

    def check_tokens(self, feature_name, value, small_words, plural, capital_words):
        for i in range(len(value)):
            tokens = word_tokenize(value[i])
            # check if there are any words should be small
            tokens = [token.lower() if token.lower() in small_words.get(feature_name,
                                                                        {}) or token.lower() in small_words else token
                      for token in tokens]
            # check if there are any words should be plural
            tokens = [token + 's' if token.lower() in plural.get(feature_name, {}) or token.lower() in plural else token
                      for token in tokens]
            # check if there are any words should be capital
            value[i] = ' '.join([token.upper() if token.lower() in capital_words.get(feature_name,
                                                                                     {}) or token.lower() in capital_words else token
                                 for token in tokens])
            value[i] = value[i].replace(" %", "%").replace(" /", "/").replace(" 路", ",")

        return value

    def clean(self, feature_name, featue_value, use_convert=False):
        # value processing
        value = self.characters_cleaning(featue_value)

        # extract domantions from 'dimensions' feature name
        value = self.get_dimensions(value) if 'dimensions' in feature_name.lower() else value

        # tokens processing
        tokens = self.tokenizing(feature_name, value)

        # seperate digits from text with white space
        tokens = self.seperate_digits(tokens)

        # Using con_map to one to one value maping
        tokens = self.map_tokens(self.con_map, feature_name, tokens)

        # extract num if feature in numeric_features
        tokens = self.get_digits(tokens) if feature_name in self.numeric_features else tokens

        lemmatized_words = self.lemmetize(feature_name, tokens, self.get_lemma)

        # Using con_map to one to one value maping again after lemmatization
        lemmatized_words = self.map_tokens(self.con_map, feature_name, lemmatized_words)

        # return original tokens if not lemmatized_words
        if not lemmatized_words:
            lemmatized_words = tokens

        value = self.join_captelaize_nouns(lemmatized_words)

        # mapping values using value_map
        value = self.map_value(self.value_map, feature_name, value)

        # standarize values using standard_values
        fparser = self.feature_parser.get(feature_name, self.feature_parser.get('default'))
        value = fparser(feature_name, value)

        # split value of multi o/p features using special_tokonizer
        if self.feature_group.get(feature_name) in self.no_join_features:
            value = list(set(self.special_tokonizer(feature_name, value)))

        # convert non list objects to list
        if type(value) != list:
            value = [value]

        value = self.check_tokens(feature_name, value, self.small_words, self.plural, self.capital_words)

        if feature_name.lower() in self.sentence:
            value = [val.capitalize() for val in value]
        return value

    def df_clean(self, row):
        return self.clean(row['Feature Name'], row['Feature Value'])
